"""
This is a class that offers a set of functions to work with meteorological data
in txt table format, separated with spaces.

It makes use of the pandas dataframe objects in order to be more efficient


Created by Pere Munar-Adrover
email: pere.munaradrover@gmail.com

"""

import os
import sys
from tqdm import *
from scipy.interpolate import interp1d
import pandas as pd
import logging
import logging.config
from molecularprofiles.utils.observatory import (
    select_new_epochs_dataframe_density_north,
    select_new_epochs_dataframe_density_south,
    select_new_epochs_dataframe_south,
    select_new_epochs_dataframe_north,
)
from molecularprofiles.utils.grib_utils import *
from molecularprofiles.utils.meteorological_constants import (
    STD_NUMBER_DENSITY,
    STD_AIR_PRESSURE,
    STD_AIR_TEMPERATURE,
    DENSITY_SCALE_HEIGHT,
)
from molecularprofiles.utils.dataframe_ops import (
    select_dataframe_by_year,
    select_dataframe_by_month,
    select_dataframe_by_hour,
    compute_averages_std,
    avg_std_dataframe,
)
from molecularprofiles.utils.humidity import (
    compressibility,
    density_moist_air,
    molar_fraction_water_vapor,
    partial_pressure_water_vapor,
)
from molecularprofiles.utils.rayleigh import Rayleigh

ROOTDIR = os.path.dirname(os.path.abspath(__file__))
log_config_file = f"{ROOTDIR}/utils/mdps_log.conf"
logging.config.fileConfig(fname=log_config_file, disable_existing_loggers=False)
logger = logging.getLogger(__name__)


class MolecularProfile:
    def __init__(self, data_file, data_server="server", tag_name="myplots", observatory="north"):
        """
        This class provides a series of functions to analyze the quality of the data for both
        CTA-North and CTA-South.

        :param data_file: txt file containing the data (string)
        :param tag_name: name to be given to the output files (string)
        :param data_server: label to be put in some of the output plots (string)
        :param observatory: valid options are: "north", "south", "other"

        Methods within this class:

        get_data:                   it retrieves the data from the input file. If the input file
                                    is a grib file and there is no file in the working directory
                                    with the same name but with .txt extension the program extracts
                                    the data from the grib file through the grib_utils program. If
                                    there is such a txt file, it reads it directly
        plot_moist_dry_comparison:  plots a comparison between the number density (scaled by an exponential)
                                    without the contribution of the humidity and the density that has
                                    taken into account the effect of the humidity
        plot_density_at_15km:       plots the number density of molecules at 15 km, where the changes due to
                                    seasonal variations are at maximum
        print_to_text_file:         prints the data into a txt file
        write_corsika:              pints the data into a txt file which format is compliant with the input card
                                    for the CORSIKA air shower simulation software

        """

        self.data_file = data_file
        self.data_server = data_server
        self.observatory = observatory
        self.tag_name = tag_name + "_" + self.data_server + "_" + self.observatory

        """
        self.molecular_n_density = (
            molecular_n_density  # [cm^-3] molecular number density for standard air conditions
        )
        self.standard_air_pressure = standard_air_pressure  # [hPa]   standard air pressure
        self.std_air_temp = std_air_temp  # [K]     standard air temperature
        DENSITY_SCALE_HEIGHT = (
            density_scale_height  # [km]    density scale hight for La Palma Winter
        )
        """

        # ----------------------------------------------------------------------

        # P = np.array([1000, 975, 950, 925, 900, 875, 850, 825, 800, 775, 750, 700, 650, 600, 550, 500, 450, 400,
        #                    350, 300, 250, 225, 200, 175, 150, 125, 100, 70, 50, 30, 20, 10, 7, 5, 3, 2, 1])

    # =============================================================================================================
    # Private functions
    # =============================================================================================================
    def _interpolate_simple(self, x_param, y_param, new_x_param):
        func = interp1d(x_param, y_param, kind="cubic", fill_value="extrapolate")
        return func(new_x_param)

    def _interpolate_param_to_h(self, param, height):
        interpolated_param = []
        group_mjd = self.dataframe.groupby("MJD")

        logger.info("Computing the extrapolation of the values of density:")
        logger.info("(This is to make it easier to compare ECMWF and GDAS, or any other")
        logger.info("weather model)")
        pbar = tqdm(total=len(self.dataframe.MJD.unique()))

        for mjd in self.dataframe.MJD.unique():
            pbar.update(1)
            h_at_mjd = group_mjd.get_group(mjd)["h"].tolist()
            param_at_mjd = group_mjd.get_group(mjd)[param].tolist()
            func = interp1d(h_at_mjd, param_at_mjd, kind="cubic", fill_value="extrapolate")

            if isinstance(height, int) or isinstance(height, float):
                interpolated_param.append(np.float(func(height)))
            else:
                interpolated_param.append(func(height))
        pbar.close()
        interpolated_param = np.array(interpolated_param)
        if isinstance(height, float) or isinstance(height, int):
            interpolated_param = np.array(interpolated_param)
            return interpolated_param
        else:
            interpolated_param_avgs = compute_averages_std(interpolated_param)
            return (
                interpolated_param,
                interpolated_param_avgs[0],
                interpolated_param_avgs[1],
                interpolated_param_avgs[2],
                interpolated_param_avgs[3],
            )

    def _compute_mass_density(self, air="moist", co2_concentration=415, interpolate=False):
        """
        Uses the functions density_moist_air, molar_fraction_water_vapor and compressibility
        from the LIDAR_analysis module humidity.py
        input:
            air: (str, optional) must be 'moist' or 'dry'

        :return: density [kg/m^3], std_dev(density) [kg/m^3], peak2peak_minus, peak2peak_plus
        """
        # C :  CO2 average global concentration in ppm taken from Keeling curve
        C = co2_concentration
        rho_s = 1.225  # kg/m^3 standard air mass density (sea level, 15 degrees C)
        rho = []

        pbar = tqdm(total=len(self.dataframe.P))
        for i in np.arange(len(self.dataframe.P)):
            pbar.update(1)
            if air == "moist":
                Xw = molar_fraction_water_vapor(
                    self.dataframe.P.iloc[i], self.dataframe.Temp.iloc[i], self.dataframe.RH.iloc[i]
                )
                Z = compressibility(self.dataframe.P.iloc[i], self.dataframe.Temp.iloc[i], Xw)
                rho.append(
                    density_moist_air(
                        self.dataframe.P.iloc[i] * 100.0, self.dataframe.Temp.iloc[i], Z, Xw, C
                    )
                )

            elif air == "dry":
                Z = compressibility(self.dataframe.P.iloc[i], self.dataframe.Temp.iloc[i], 0.0)
                rho.append(
                    density_moist_air(
                        self.dataframe.P.iloc[i] * 100.0, self.dataframe.Temp.iloc[i], Z, 0.0, C
                    )
                )
            else:
                logger.critical('Wrong air condition. It must be "moist" or "dry". Aborting!')
                sys.exit()
        pbar.close()

        self.dataframe["n_mass_" + air] = rho
        self.dataframe["nexp_mass_" + air] = (
            self.dataframe["n_mass_" + air]
            / rho_s
            * np.exp(self.dataframe.h / DENSITY_SCALE_HEIGHT)
        )

    def _compare_models(self, model_1, model_2):
        """
        Compares two atmospheric models and returns statistical difference between them
        """
        pass
        """
        Reference content of removed similar function:
        diff_with_magic = []
        diff_with_prod3 = []

        interpolated_density = self._interpolate_param_to_h("n_exp", self.x)[0]

        self._get_prod3sim_data()
        x = np.linspace(1000.0, 25000.0, num=15, endpoint=True)

        logger.info("Computing the differences of the values of density:")
        for i in tqdm(np.arange(len(interpolated_density))):
            diff_with_magic.append(
                (interpolated_density[i] - self.func_magic(x)) / interpolated_density[i]
            )
            diff_with_prod3.append(
                (interpolated_density[i] - self.func_prod3(x)) / interpolated_density[i]
            )

        self.diff_with_magic = np.asarray(diff_with_magic)
        self.diff_with_prod3 = np.asarray(diff_with_prod3)
        logger.info("DIFF PROD3", self.diff_with_prod3)
        self.diff_MAGIC = compute_averages_std(self.diff_with_magic)
        self.diff_PROD3 = compute_averages_std(self.diff_with_prod3)
        """

    # =============================================================================================================
    # Main get data function
    # =============================================================================================================
    def get_data(
        self,
        epoch="all",
        years=None,
        months=None,
        hours=None,
        altitude=[],
        select_good_weather=False,
        RH_lim=100.0,
        W_lim=10000.0,
        filter_by_density=False,
        n_exp_minvalue=0.0,
        n_exp_maxvalue=1.0,
        filter_by_density_name="winter",
    ):

        """
        Function that reads ECMWF or GDAS txt input files and returns quantities ready to plot.
        If the input filename does not exist, the program searches for the same
        input name with .grib extension and extracts the data from it

        Input: epoch:  (str) can be "winter", "summer", "intermediate", "all". Default is "all"
               years:  (list) a sequence of years to filter the data
               months: (list) a sequence of months to filter the data (elements values must be
                       between 1 and 12)
               hours:  (list) a sequence of hours to filter the data (elements values must be
                       either 0,6,12 or 18)
               altitude: (2-element list) boundaries for the minimum and maximum altitude, in meters, between
                         which the final dataframe will be filtered
               select_good_weather: (bool) if True it uses RH_lim and W_lim parameters to filter the
                                    data by good weather conditions
               RH_lim: (float) if select_good_weather is True, this is the RH upper limit value to use
                       to filter the data
               W_lim: (float) if select_good_weather is True, this is the Wind speed upper limit value
                      to use to filter the data
               filter_by_density: (bool) if set to True it filters the data by density values instead of
                                  filtering by epoch

        :return:
            self.dataframe
            self.h_avgs
            self.n_avgs
            self.Temp_avgs
            self.wind_speed_avgs
            self.wind_direction_avgs
            self.RH_avgs
            self.n_exp_avgs
        """

        if not os.path.exists(os.path.splitext(self.data_file)[0] + ".txt"):
            grib_file = self.data_file
            if self.data_server == "GDAS":
                gridstep = 1.0
            elif self.data_server == "ECMWF":
                gridstep = 0.75
            else:
                gridstep = 0.75
            read_grib_file_to_text(grib_file, gridstep, self.observatory)

        self.output_plot_name = self.tag_name + "_" + epoch
        self.epoch = epoch

        self.dataframe = pd.read_csv(
            os.path.splitext(self.data_file)[0] + ".txt", sep=" ", comment="#"
        )
        self.dataframe["n_exp"] = (
            self.dataframe.n / STD_NUMBER_DENSITY * np.exp(self.dataframe.h / DENSITY_SCALE_HEIGHT)
        )

        # Altitude filtering:
        if altitude != [] and len(altitude) == 2 and altitude[0] < altitude[1]:
            altitude_cond = (self.dataframe.h >= altitude[0]) & (self.dataframe.h < altitude[1])
            self.dataframe = self.dataframe[altitude_cond]
            self.x = np.linspace(altitude[0], altitude[1], num=15, endpoint=True)
        elif altitude != [] and len(altitude) == 1:
            altitude_cond = self.dataframe.h < altitude[0]
            self.dataframe = self.dataframe[altitude_cond]
            self.x = np.linspace(2200.0, altitude[0], num=15, endpoint=True)
        elif altitude == []:
            self.x = np.linspace(2200.0, 25000.0, num=15, endpoint=True)
        elif len(altitude) > 2 or altitude[0] > altitude[1]:
            logger.critical("bad altitude filter. Aborting")
            sys.exit()

        # Filtering by years or months or hours
        if years:
            self.dataframe = select_dataframe_by_year(self.dataframe, years)
        if months:
            self.dataframe = select_dataframe_by_month(self.dataframe, months)
        if hours:
            self.dataframe = select_dataframe_by_hour(self.dataframe, hours)

        # Filtering by epoch
        if epoch != "all" and not years and not months:
            if self.observatory == "north":
                self.dataframe = select_new_epochs_dataframe_north(self.dataframe, epoch)
            elif self.observatory == "south":
                self.dataframe = select_new_epochs_dataframe_south(self.dataframe, epoch)
            else:
                self.dataframe = select_new_epochs_dataframe_north(self.dataframe, epoch)

        # Filtering by n_exp density value
        elif filter_by_density:
            if epoch != "all":
                if self.observatory == "north":
                    self.dataframe = select_new_epochs_dataframe_density_north(
                        self.dataframe, epoch
                    )
                elif self.observatory == "south":
                    self.dataframe = select_new_epochs_dataframe_density_south(
                        self.dataframe, epoch
                    )
                else:
                    self.dataframe = select_new_epochs_dataframe_density_north(
                        self.dataframe, epoch
                    )
            self.epoch = filter_by_density_name

        # Filtering by good weather conditions:
        if select_good_weather:
            self.dataframe["W"] = np.sqrt(self.dataframe.U**2.0 + self.dataframe.V**2.0)
            h_cond = (self.dataframe.h > 2200.0) & (self.dataframe.h < 2500.0)
            gw_cond = (self.dataframe.RH < RH_lim) & (self.dataframe.W < W_lim * 1000.0 / 3600.0)
            self.dataframe["datehour"] = self.dataframe.Date.astype(
                str
            ) + self.dataframe.hour.astype(str)
            good_weather_dates = self.dataframe.datehour[(h_cond) & (gw_cond)]
            self.dataframe = self.dataframe[
                (self.dataframe.datehour.isin(good_weather_dates.tolist()))
            ]

        # Various averaged values obtained after filtering (if no filter, averages are made over the whole dataframe)
        self.group_by_p = self.dataframe.groupby("P")
        self.h_avgs = avg_std_dataframe(self.group_by_p, "h")
        self.n_avgs = avg_std_dataframe(self.group_by_p, "n")
        self.Temp_avgs = avg_std_dataframe(self.group_by_p, "Temp")
        self.wind_speed_avgs = avg_std_dataframe(self.group_by_p, "wind_speed")
        self.wind_direction_avgs = avg_std_dataframe(self.group_by_p, "wind_direction")
        self.RH_avgs = avg_std_dataframe(self.group_by_p, "RH")
        self.n_exp_avgs = avg_std_dataframe(self.group_by_p, "n_exp")

    # =======================================================================================================
    # printing functions:
    # =======================================================================================================
    def print_to_text_file(self):
        textfile = open(self.output_plot_name + "_to_text_file.txt", "w")
        print("# x[i], avg, rms, p2p_m, p2p_p", file=textfile)
        for i in np.arange(len(self.x)):
            print(
                self.x[i],
                self.averages[0][i],
                self.averages[1][i],
                self.averages[3][i],
                self.averages[2][i],
                self.diff_MAGIC[0][i],
                self.diff_MAGIC[1][i],
                self.diff_MAGIC[3][i],
                self.diff_MAGIC[2][i],
                file=textfile,
            )
        textfile.close()

    def _refractive_index(self, P, T, RH, wavelength, C):
        """Wrapper for Rayleigh.calculate_n()."""
        rayleigh = Rayleigh(wavelength, C, P, T, RH)
        return rayleigh.calculate_n()

    def write_corsika(self, outfile, co2_concentration):
        """
        Write an output file in the style of a CORSIKA atmospheric configuration file:

        alt (km)     rho (g/cm^3)   thick (g/cm^2)   (n-1)
        """
        mbar2gcm2 = 1.019716213  # conversion from mbar (pressure in SI) to g/cm^2 (pressure in cgs)
        # Loschmidt constant: number density of particles in an ideal gas at STP in m^-3
        N0 = 2.079153e25  # Na divided by molar mass of air: 0.0289644
        height = np.array(
            [
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
                6.0,
                7.0,
                8.0,
                9.0,
                10.0,
                11.0,
                12.0,
                13.0,
                14.0,
                15.0,
                16.0,
                17.0,
                18.0,
                19.0,
                20.0,
                21.0,
                22.0,
                23.0,
                24.0,
                25.0,
                26.0,
            ]
        )

        with open(outfile, "w") as f:

            f.write("# Atmospheric Model ECMWF year/month/day   hour h\n")  # .format(**datedict))
            f.write(
                "#Col. #1          #2           #3            #4        [ #5 ]        [ #6 ]       [ # 7 ]\n"
            )
            f.write(
                "# Alt [km]    rho [g/cm^3] thick [g/cm^2]    n-1        T [k]       p [mbar]      pw / p\n"
            )

            density = self.n_avgs[0].sort_index(ascending=False).values
            density /= N0 * 1.0e-3
            P = self.dataframe.P.unique()[::-1]

            # height = self.h_avgs[0].sort_index(ascending=False).values / 1.e3
            # Temp = self.Temp_avgs[0].sort_index(ascending=False).values
            # RH = self.RH_avgs[0].sort_index(ascending=False).values

            T0 = float(
                self._interpolate_simple(
                    self.h_avgs[0].sort_index(ascending=False).values / 1e3,
                    self.Temp_avgs[0].sort_index(ascending=False).values,
                    0.0,
                )
            )
            RH0 = float(
                self._interpolate_simple(
                    self.h_avgs[0].sort_index(ascending=False).values / 1e3,
                    self.RH_avgs[0].sort_index(ascending=False).values,
                    0.0,
                )
            )
            P0 = float(
                self._interpolate_simple(
                    self.h_avgs[0].sort_index(ascending=False).values / 1e3, P, 0.0
                )
            )
            density0 = float(
                self._interpolate_simple(
                    self.h_avgs[0].sort_index(ascending=False).values / 1e3, density, 0.0
                )
            )
            thick0 = P0 * mbar2gcm2

            pw0 = partial_pressure_water_vapor(T0, RH0) / P0

            Temp = self._interpolate_simple(
                self.h_avgs[0].sort_index(ascending=False).values / 1.0e3,
                self.Temp_avgs[0].sort_index(ascending=False).values,
                height,
            )
            RH = self._interpolate_simple(
                self.h_avgs[0].sort_index(ascending=False).values / 1.0e3,
                self.RH_avgs[0].sort_index(ascending=False).values,
                height,
            )
            RH[RH < 0.0] = 1.0e-4
            Pressure = self._interpolate_simple(
                self.h_avgs[0].sort_index(ascending=False).values / 1.0e3, P, height
            )
            density = self._interpolate_simple(
                self.h_avgs[0].sort_index(ascending=False).values / 1.0e3, density, height
            )
            thick = Pressure * mbar2gcm2
            pwp = partial_pressure_water_vapor(Temp, RH) / Pressure

            nm0 = self._refractive_index(P0, T0, RH0, 350.0, co2_concentration) - 1.0
            outdict = {
                "height": 0.000,
                "rho": density0,
                "thick": thick0,
                "nm1": nm0,
                "T": T0,
                "p": P0,
                "pw/p": pw0,
            }
            f.write(
                "  {height:7.3f}     {rho:5.5E}  {thick:5.5E}  {nm1:5.5E}  {T:5.5E}  {p:5.5E}  {pw/p:5.5E}\n".format(
                    **outdict
                )
            )

            for i in np.arange(len(height)):
                nm1 = (
                    self._refractive_index(Pressure[i], Temp[i], RH[i], 350, co2_concentration) - 1
                )

                outdict = {
                    "height": height[i],
                    "rho": density[i],
                    "thick": thick[i],
                    "nm1": nm1,
                    "T": Temp[i],
                    "p": Pressure[i],
                    "pw/p": pwp[i],
                }
                f.write(
                    "  {height:7.3f}     {rho:5.5E}  {thick:5.5E}  {nm1:5.5E}  {T:5.5E}  {p:5.5E}  {pw/p:5.5E}\n".format(
                        **outdict
                    )
                )

            # concatenate the dummy values of MagicWinter starting from 50.0 km
            f.write(
                "   27.000     2.96567e-05  1.96754e+01  6.84513e-06  223.70  1.90428e+01  0.00000e+00\n"
            )

            f.write(
                "   28.000     2.52913e-05  1.69341e+01  5.83752e-06  225.69  1.63842e+01  0.00000e+00\n"
            )

            f.write(
                "   29.000     2.15935e-05  1.45949e+01  4.98403e-06  227.75  1.41164e+01  0.00000e+00\n"
            )

            f.write(
                "   30.000     1.84554e-05  1.25967e+01  4.25972e-06  229.92  1.21798e+01  0.00000e+00\n"
            )

            f.write(
                "   32.000     1.35171e-05  9.42594e+00  3.11990e-06  234.75  9.10824e+00  0.00000e+00\n"
            )

            f.write(
                "   34.000     9.93769e-06  7.09977e+00  2.29374e-06  240.36  6.85638e+00  0.00000e+00\n"
            )

            f.write(
                "   36.000     7.35111e-06  5.38483e+00  1.69672e-06  246.30  5.19717e+00  0.00000e+00\n"
            )

            f.write(
                "   38.000     5.47808e-06  4.11190e+00  1.26441e-06  252.24  3.96630e+00  0.00000e+00\n"
            )

            f.write(
                "   40.000     4.11643e-06  3.15961e+00  9.50121e-07  257.79  3.04599e+00  0.00000e+00\n"
            )

            f.write(
                "   42.000     3.12126e-06  2.44097e+00  7.20424e-07  262.51  2.35189e+00  0.00000e+00\n"
            )

            f.write(
                "   44.000     2.38908e-06  1.89362e+00  5.51429e-07  265.91  1.82354e+00  0.00000e+00\n"
            )

            f.write(
                "   46.000     1.84598e-06  1.47276e+00  4.26075e-07  267.52  1.41753e+00  0.00000e+00\n"
            )

            f.write(
                "   48.000     1.43779e-06  1.14626e+00  3.31858e-07  267.20  1.10275e+00  0.00000e+00\n"
            )

            f.write(
                "   50.000     1.09738e-06  8.72656e-01   2.53289e-07   266.34   8.38955e-01  0.00000e+00\n"
            )
            f.write(
                "   55.000     5.99974e-07  4.61036e-01   1.38481e-07   257.19   4.42930e-01  0.00000e+00\n"
            )
            f.write(
                "   60.000     3.25544e-07  2.36175e-01   7.51395e-08   242.81   2.26896e-01  0.00000e+00\n"
            )
            f.write(
                "   65.000     1.70152e-07  1.15918e-01   3.92732e-08   227.93   1.11324e-01  0.00000e+00\n"
            )
            f.write(
                "   70.000     8.43368e-08  5.45084e-02   1.94660e-08   215.90   5.22651e-02  0.00000e+00\n"
            )
            f.write(
                "   75.000     3.95973e-08  2.48012e-02   9.13953e-09   208.66   2.37169e-02  0.00000e+00\n"
            )
            f.write(
                "   80.000     1.79635e-08  1.10899e-02   4.14618e-09   205.11   1.05760e-02  0.00000e+00\n"
            )
            f.write(
                "   85.000     8.03691e-09  4.91583e-03   1.85502e-09   202.12   4.66284e-03  0.00000e+00\n"
            )
            f.write(
                "   90.000     3.59602e-09  2.15599e-03   8.30003e-10   196.26   2.02583e-03  0.00000e+00\n"
            )
            f.write(
                "   95.000     1.59871e-09  9.21029e-04   3.69000e-10   187.55   8.60656e-04  0.00000e+00\n"
            )
            f.write(
                "  100.000     6.73608e-10  3.82814e-04   1.55477e-10   185.38   3.58448e-04  0.00000e+00\n"
            )
            f.write(
                "  105.000     2.69097e-10  1.61973e-04   6.21108e-11   197.19   1.52311e-04  0.00000e+00\n"
            )
            f.write(
                "  110.000     1.09021e-10  7.37110e-05   2.51634e-11   224.14   7.01416e-05  0.00000e+00\n"
            )
            f.write(
                "  115.000     4.71300e-11  3.70559e-05   1.08782e-11   268.51   3.63251e-05  0.00000e+00\n"
            )
            f.write(
                "  120.000     2.23479e-11  2.05900e-05   5.15817e-12   333.43   2.13890e-05  0.00000e+00\n"
            )

    def create_mdp(self, mdp_file):
        """
        Write an output file with the molecular number density per height
        """

        height = np.arange(
            1.0, 27.0, 1
        )  # FIXME: The hardcoded value 27 reflects the current ceiling of GDAS data (26km a.s.l.). Shouldn't be hardcoded and in general the binning and limits should be considered.
        with open(mdp_file, "w") as f:

            number_density_exp = self._interpolate_simple(
                self.h_avgs[0].sort_index(ascending=False).values / 1.0e3,
                self.n_exp_avgs[0].sort_index(ascending=False).values,
                height,
            )

            for i in np.arange(len(height)):
                file_line = [str(height[i]), "\t", str(number_density_exp[i]), "\n"]
                f.writelines(file_line)



Ns = 2.546899e19  # [cm^-3] molecular number density for standard air conditions
ps = 1013.25  # [hPa]   standard air pressure
Ts = 288.15  # [K]     standard air temperature
Hs = 9500.0  # [km]    density scale hight for La Palma Winter



"""
This module contains functions that operate in the dataframe
created based on GDAS/ECMWF analysis data
"""
import pandas as pd
import numpy as np
from statsmodels import robust
from molecularprofiles.utils.observatory import get_epoch


def select_dataframe_epoch(dataframe_atmo_profile, epoch_text):
    """Select data for a specific epoch"""
    epoch = get_epoch(epoch_text)
    new_dataframe_atmo_profile = dataframe_atmo_profile[dataframe_atmo_profile.month.isin(epoch)]
    return new_dataframe_atmo_profile


def select_dataframe_by_year(dataframe_atmo_profile, years):
    """Select data for specific years"""
    new_dataframe_atmo_profile = dataframe_atmo_profile[dataframe_atmo_profile.year.isin(years)]
    return new_dataframe_atmo_profile


def select_dataframe_by_month(dataframe_atmo_profile, months):
    """Select data for specific months"""
    new_dataframe_atmo_profile = dataframe_atmo_profile[dataframe_atmo_profile.month.isin(months)]
    return new_dataframe_atmo_profile


def select_dataframe_by_hour(dataframe_atmo_profile, hours):
    """Select data for specific hours"""
    new_dataframe_atmo_profile = dataframe_atmo_profile[dataframe_atmo_profile.hour.isin(hours)]
    return new_dataframe_atmo_profile


def create_wind_speed_dataframe(dataframe_atmo_profile, normalized=False):
    """
    Function to create a wind speed dataframe in order to plot it afterwards as a wind rose plot

    :param dataframe_atmo_profile: dataframe containing wind direction information
    :param normalized (optional):
    :return: dataframe_atmo_profile_winds
    """
    wd_centre_bins = np.arange(7.5, 360, 15)
    ws_hist = []
    for d in wd_centre_bins:
        ws_hist.append(
            np.histogram(
                dataframe_atmo_profile.wind_speed[
                    (dataframe_atmo_profile.wind_direction >= d - 7.5)
                    & (dataframe_atmo_profile.wind_direction < d + 7.5)
                ],
                bins=[0, 5, 10, 20, 30, 40, 50, 100],
            )[0]
        )

    dataframe_atmo_profile_winds = pd.DataFrame(
        columns=["wind_direction", "0-5", "5-10", "10-20", "20-30", "30-40", "40-50", "> 50"]
    )
    ws_new_list = []
    for j in range(len(ws_hist[0])):
        li = []
        for i in range(len(ws_hist)):
            li.append(ws_hist[i][j])
        ws_new_list.append(li)

    for i, j in zip(dataframe_atmo_profile_winds.keys()[1:], range(len(ws_new_list))):
        dataframe_atmo_profile_winds[i] = ws_new_list[j]
    if normalized:
        dataframe_atmo_profile_winds_normalized = dataframe_atmo_profile_winds.div(
            dataframe_atmo_profile_winds.sum(axis=1), axis=0
        )
        dataframe_atmo_profile_winds_normalized["wind_direction"] = wd_centre_bins
        return dataframe_atmo_profile_winds_normalized
    else:
        dataframe_atmo_profile_winds["wind_direction"] = wd_centre_bins
        return dataframe_atmo_profile_winds


def compute_averages_std_simple(input_array):
    """
    This function computes the average, standard deviation and peak to peak (plus and minus)
    for an input 1D array

    Input:
        1-D input_array (array-like)

    Output:
        average, stand_dev, peak_to_peak_p, peak_to_peak_m (array-like)
    """

    average = np.average(input_array)
    stand_dev = robust.mad(input_array)
    peak_to_peak_p = np.max(input_array) - np.average(input_array)
    peak_to_peak_m = np.average(input_array) - np.min(input_array)
    return average, stand_dev, peak_to_peak_p, peak_to_peak_m


def avg_std_dataframe(group, param):
    """

    :param group: dataframe grouped by a certain parameter
    :param param: the parameter by which the dataframe is grouped
    :return:
        avg: the mean value for each grouped level
        std: the standard deviation for each grouped level
        mad: the mean absolute deviation for each group level
        p2p_p: the peak-to-peak maximum value for each grouped level
        p2p_m: the peak-to-peak minimum value for each grouped level
    """

    avg = group[param].mean()
    std = group[param].std()
    mad = group[param].mad()
    p2p_p = group[param].max() - avg
    p2p_m = avg - group[param].min()

    return avg, std, mad, p2p_p, p2p_m


def compute_averages_std(input_array):
    """
    This function computes the average, standard deviation and peak to peak (plus and minus)
    for a multidimensional input array

    Input: input_array (array-like)

    Output: average, stand_dev, peak_to_peak_p, peak_to_peak_m (all array-like)
    """

    average = []
    stand_dev = []
    peak_to_peak_p = []
    peak_to_peak_m = []

    if len(np.shape(input_array)) == 1:
        average.append(np.average(input_array))
        stand_dev.append(robust.mad(input_array))
        peak_to_peak_p.append(np.max(input_array - np.average(input_array)))
        peak_to_peak_m.append(np.average(input_array) - np.min(input_array))
    else:
        r = len(input_array[0])
        for i in np.arange(r):
            average.append(np.average(input_array[:, i]))
            stand_dev.append(robust.mad(input_array[:, i]))
            peak_to_peak_p.append(np.max(input_array[:, i] - np.average(input_array[:, i])))
            peak_to_peak_m.append(np.average(input_array[:, i]) - np.min(input_array[:, i]))

    average = np.asarray(average)
    stand_dev = np.asarray(stand_dev)
    peak_to_peak_p = np.asarray(peak_to_peak_p)
    peak_to_peak_m = np.asarray(peak_to_peak_m)
    return average, stand_dev, peak_to_peak_p, peak_to_peak_m



#!/usr/bin/python3

"""
set of utilities for decoding & analysing grib files
"""

from builtins import str
import sys
import gc
import argparse
import os
import logging
import logging.config
import multiprocessing
from multiprocessing import Process
import numpy as np
import pygrib as pg
from molecularprofiles.utils.observatory import get_observatory_coordinates, get_closest_gridpoint
from tqdm import tqdm
from astropy.time import Time
from molecularprofiles.utils.meteorological_constants import (
    STD_NUMBER_DENSITY,
    STD_AIR_PRESSURE,
    STD_AIR_TEMPERATURE,
    DENSITY_SCALE_HEIGHT,
)

ROOTDIR = os.path.dirname(os.path.abspath(__file__))
log_config_file = f"{ROOTDIR}/mdps_log.conf"
logging.config.fileConfig(fname=log_config_file, disable_existing_loggers=False)
logger = logging.getLogger(__name__)


def get_altitude_from_geopotential(geop_height, latitude_obs):
    """
    Function to compute the real altitude from the geopotential value at
    a certain coordinates on Earth
    :param geop_height:
    :param latitude_obs: geographical latitude of interest in degrees
    :return: real altitude as fGeoidOffset (in m)
    """
    latitude = np.radians(latitude_obs)
    geop_heightkm = (
        geop_height / 1000.0 / 9.80665
    )  # dividing by the acceleration of gravity on Earth
    cos2lat = np.cos(2 * latitude)
    # convert from geopotential height to geometric altitude:
    # Uses expression 20 from http://www.earth.sinica.edu.tw/~bfchao/publication/eng/2005-Boy&Chao_JGR2005_precise%20evaluation%20of%20atmospheric%20loading%20effects%20on%20earths%20time-variable%20gravity%20field.pdf
    geo_altitude = (1.0 + 0.002644 * cos2lat) * geop_heightkm + (
        1 + 0.0089 * cos2lat
    ) * geop_heightkm * geop_heightkm / 6245.0
    # convert geometric altitude to meter
    return 1.0e3 * geo_altitude  # This is fGeoidOffset


def get_altitude_from_geopotential_height(geop, latitude_obs):
    """
    Function to compute the real altitude from the geopotential value at
    a certain coordinates on Earth
    :param geop_height:
    :param latitude_obs: geographical latitude of interest in degrees
    :return: real altitude as fGeoidOffset (in m)
    """
    latitude = np.radians(latitude_obs)
    geop_km = geop / 1000.0
    cos2lat = np.cos(2 * latitude)
    # convert from geopotential height to geometric altitude:
    # Uses expression 20 from http://www.earth.sinica.edu.tw/~bfchao/publication/eng/2005-Boy&Chao_JGR2005_precise%20evaluation%20of%20atmospheric%20loading%20effects%20on%20earths%20time-variable%20gravity%20field.pdf
    geoid_offset = (1.0 + 0.002644 * cos2lat) * geop_km + (
        1 + 0.0089 * cos2lat
    ) * geop_km * geop_km / 6245.0
    # convert GEOID_OFFSET to meter
    return 1.0e3 * geoid_offset  # This is fGeoidOffset


def date2mjd(year, month, day, hour, minute=0, second=0):
    """
    This function computes the mjd value corresponding to an input date, in the year, month, day, hour format

    Input:
        (integers) year, month, day and hour.
        Optionals: minutes and seconds. If left blank, they are assumed equal 0

    Output:
        (float) mjd
    """

    year = str(int(year))
    month = f"{int(month):02d}"
    day = f"{int(day):02d}"
    hour = f"{int(hour):02d}"
    minute = f"{int(minute):02d}"
    second = f"{int(second):02d}"

    t = Time(
        year + "-" + month + "-" + day + "T" + hour + ":" + minute + ":" + second,
        format="isot",
        scale="utc",
    )
    return round(t.mjd, 2)


def mjd2date(mjd):
    """
    This function computes the date from an input mjd day

     Input: mjd (float)

    Output: year, month, day, hour (integers) corresponding to the mjd input day
    """

    time = Time(mjd, format="mjd")
    return (
        int(time.datetime.year),
        int(time.datetime.month),
        int(time.datetime.day),
        int(time.datetime.hour),
    )


def get_gribfile_variables(file_name):
    """
    Function that returns all the different variable names in a grib file
    :param file_name:
    :return: varname (list): variable names
             varshortname (list): variable short names
    """
    logger.info("opening file...")
    with pg.open(file_name) as grib_file:
        varshortname = []
        varname = []
        grib_file.read(10)[0]
        while True:
            v = grib_file.read(1)[0]
            variable_name = v.name.replace(" ", "")
            variable_short_name = v.shortName
            if variable_name not in varname:
                varname.append(variable_name)
                varshortname.append(variable_short_name)
            else:
                break
    return varname, varshortname


def get_plevels(atmo_parameter):
    """
    The measurements of atmospheres parameters are taking place in various pressure levels.
    This functions returns the indices of these levels where we have measurements
    for the given atmospheric parameter.
    e.g. for GDAS data that should be a range of values [1, 23]

    :param atmo_parameter: Atmospheric parameter from grib file.
    :type atmo_parameter: string

    :return: list with the indices of the pressure levels
    """

    plevels = []
    index = []
    for i in range(len(atmo_parameter)):
        pressure_level = atmo_parameter[i].level
        if pressure_level not in plevels:
            plevels.append(pressure_level)
            index.append(i + 1)
        else:
            break
    return index


def get_grib_file_data(file_name):
    """
    This function opens a grib file, selects the parameters
    (all available: Temperature, Geopotential, RH, ...),
    and creates a dictionary variable with these parameters.
    It also returns the variable name and variable short name

    Input: file_name (string)
           observatory (string). Possible values are 'north' or 'south'
           gridstep (float): grid spacing in degrees.
           Values are 1.0 for GDAS data and 0.75 for ECMWF data.

    Output: a txt file with the exact name as the input file name, but with .txt as extension
    """

    logger.info(f"Working on {file_name}")
    logger.info("getting all variable names...")
    variable_name, variable_short_name = get_gribfile_variables(file_name)
    logger.info(f"indexing the file {file_name} (this might take a while...)")
    grib_file = pg.index(file_name, "shortName", "typeOfLevel")
    logger.info(
        f"selecting the parameters information for {file_name} (this might take a while...)"
    )
    data = []

    for short_name in variable_short_name:
        var = grib_file.select(shortName=short_name, typeOfLevel="isobaricInhPa")
        data.append(var)
        gc.collect()

    datadict = dict(zip(variable_name, data))
    data = None
    gc.collect()
    return variable_name, datadict


def fill_relative_humidity_gaps(rhdata):
    """in case the RH data for Plevels 20 and 50 hPa is not present, this function fills these gaps
    with 0.0"""
    relative_humidity = []
    for i in np.arange(len(rhdata)):
        relative_humidity.append(rhdata[i].values)
    relative_humidity = np.asarray(relative_humidity)

    for i in np.arange(len(relative_humidity)):
        if i * 23 < len(relative_humidity):
            relative_humidity = np.insert(relative_humidity, (i * 23, i * 23), 0.0)
    return relative_humidity


def compute_density(pressure, temperature):
    """Return molecular number density"""
    return STD_NUMBER_DENSITY * pressure / STD_AIR_PRESSURE * STD_AIR_TEMPERATURE / temperature


def compute_wind_direction(u_direction, v_direction):
    """Return wind direction"""
    angle = np.arctan2(-1 * u_direction, -1 * v_direction) * 180.0 / np.pi
    if angle < 0.0:
        angle += 360.0
    direction = angle
    return direction


def compute_wind_speed(u_direction, v_direction):
    """Return wind speed"""
    return np.sqrt(u_direction**2.0 + v_direction**2.0)


def read_grib_file_to_text(file_name, gridstep, observatory=None, lat=None, lon=None):
    """
    This function creates a txt file where the information from the
    get_grib_file_data function is written, together with date, year,
    month, day, hour, pressure level, real height and density.

    Input: file_name (string)
           observatory (string). Possible values are 'north', 'south'
           or any other name. If NOT north or south,
           then the program asks for the coordinates.
           gridstep (float): grid spacing in degrees. Values are 1.0 for GDAS data
           and 0.75 for ECMWF data.
           lat: (float, optional) latitude of the observatory in degrees
           lon: (float, optional) longitude of the observatory in degrees
    Output: a txt file with the exact name as the input file name,
            but with .txt as extension
    """

    if os.path.exists(os.path.splitext(file_name)[0] + ".txt"):
        logger.critical(
            f"Output file {os.path.splitext(file_name)[0]}.txt already exists. Aborting."
        )
        sys.exit()

    vn, datadict = get_grib_file_data(file_name)
    latitude_obs, longitude_obs = None, None

    if observatory:
        latitude_obs, longitude_obs = get_observatory_coordinates(observatory)
    elif lat and lon:
        latitude_obs, longitude_obs = lat, lon

    lat_gridpoint, lon_gridpoint = get_closest_gridpoint(latitude_obs, longitude_obs, gridstep)

    if len(datadict["Temperature"]) != len(datadict["Relativehumidity"]):
        relative_humidity = fill_relative_humidity_gaps(datadict["Relativehumidity"])
    else:
        relative_humidity = []
        for i in np.arange(len(datadict["Relativehumidity"])):
            relative_humidity.append(datadict["Relativehumidity"][i].values)
        relative_humidity = np.asarray(relative_humidity)

    # We create the table file and fill it with the information stored in the above variables,
    # plus the height and density computed form them.

    logger.info("creating the txt file containing the selected data...")
    table_file = open(file_name.split(".")[0] + ".txt", "w")
    print(
        "Date year month day hour MJD P Temp h n n_exp U V wind_speed wind_direction RH",
        file=table_file,
    )

    pbar = tqdm(total=len(datadict["Temperature"]))
    for j in np.arange(len(datadict["Temperature"])):
        pbar.update(1)
        if (type(datadict["Temperature"][j].values) == float) or (
            len(datadict["Temperature"][j].values) == 1
        ):
            if "GeopotentialHeight" in vn:
                height = get_altitude_from_geopotential_height(
                    datadict["GeopotentialHeight"][j].values, latitude_obs
                )
            else:
                height = get_altitude_from_geopotential(
                    datadict["Geopotential"][j].values, latitude_obs
                )
            density = compute_density(
                datadict["Temperature"][j].level, datadict["Temperature"][j].values
            )
            density_exp = density / STD_NUMBER_DENSITY * np.exp(height / DENSITY_SCALE_HEIGHT)
            wind_speed = compute_wind_speed(
                datadict["Ucomponentofwind"][j].values, datadict["Vcomponentofwind"][j].values
            )
            wind_direction = compute_wind_direction(
                datadict["Ucomponentofwind"][j].values, datadict["Vcomponentofwind"][j].values
            )
            mjd = date2mjd(
                datadict["Temperature"][j].year,
                datadict["Temperature"][j].month,
                datadict["Temperature"][j].day,
                datadict["Temperature"][j].hour,
            )
            print(
                int(datadict["Temperature"][j].dataDate),
                datadict["Temperature"][j].year,
                datadict["Temperature"][j].month,
                datadict["Temperature"][j].day,
                datadict["Temperature"][j].hour,
                mjd,
                datadict["Temperature"][j].level,
                datadict["Temperature"][j].values,
                height,
                density,
                density_exp,
                datadict["Ucomponentofwind"][j].values,
                datadict["Vcomponentofwind"][j].values,
                wind_speed,
                wind_direction,
                relative_humidity[j],
                file=table_file,
            )

        else:  # this is just in case the grib file contains more than one grid point
            if "GeopotentialHeight" in vn:
                height = get_altitude_from_geopotential_height(
                    np.float(
                        datadict["GeopotentialHeight"][j].values[
                            (datadict["GeopotentialHeight"][j].data()[1] == lat_gridpoint)
                            & (datadict["GeopotentialHeight"][j].data()[2] == lon_gridpoint)
                        ]
                    ),
                    latitude_obs,
                )
            else:
                height = get_altitude_from_geopotential(
                    np.float(
                        datadict["Geopotential"][j].values[
                            (datadict["Geopotential"][j].data()[1] == lat_gridpoint)
                            & (datadict["Geopotential"][j].data()[2] == lon_gridpoint)
                        ]
                    ),
                    latitude_obs,
                )
            temperature = np.float(
                datadict["Temperature"][j].values[
                    (datadict["Temperature"][j].data()[1] == lat_gridpoint)
                    & (datadict["Temperature"][j].data()[2] == lon_gridpoint)
                ]
            )
            # Why for relative humidity,  wind speed and direction we are not
            # choosing only the grid point closest to
            # ORM, like we do for Geopotential and temperature?
            # I add the selection below
            relative_humidity = np.float(
                datadict["Relativehumidity"][j].values[
                    (datadict["Relativehumidity"][j].data()[1] == lat_gridpoint)
                    & (datadict["Relativehumidity"][j].data()[2] == lon_gridpoint)
                ]
            )
            density = compute_density(datadict["Temperature"][j].level, temperature)
            U_component_of_wind = np.float(
                datadict["Ucomponentofwind"][j].values[
                    (datadict["Ucomponentofwind"][j].data()[1] == lat_gridpoint)
                    & (datadict["Ucomponentofwind"][j].data()[2] == lon_gridpoint)
                ]
            )
            V_component_of_wind = np.float(
                datadict["Vcomponentofwind"][j].values[
                    (datadict["Vcomponentofwind"][j].data()[1] == lat_gridpoint)
                    & (datadict["Vcomponentofwind"][j].data()[2] == lon_gridpoint)
                ]
            )
            wind_speed = compute_wind_speed(U_component_of_wind, V_component_of_wind)
            wind_direction = compute_wind_direction(U_component_of_wind, V_component_of_wind)
            density_exp = density / STD_NUMBER_DENSITY * np.exp(height / DENSITY_SCALE_HEIGHT)
            mjd = date2mjd(
                datadict["Temperature"][j].year,
                datadict["Temperature"][j].month,
                datadict["Temperature"][j].day,
                datadict["Temperature"][j].hour,
            )

            print(
                int(datadict["Temperature"][j].dataDate),
                datadict["Temperature"][j].year,
                datadict["Temperature"][j].month,
                datadict["Temperature"][j].day,
                datadict["Temperature"][j].hour,
                mjd,
                datadict["Temperature"][j].level,
                temperature,
                height,
                density,
                density_exp,
                U_component_of_wind,
                V_component_of_wind,
                wind_speed,
                wind_direction,
                relative_humidity,
                file=table_file,
            )

    table_file.close()
    pbar.close()
    datadict = None


def read_grib_file_to_magic(file_name, gridstep, observatory=None, lat=None, lon=None):
    """
    This function opens a grib file, selects all parameters
    and finally creates a txt file where these parameters, together with date, year, month,
    day, hour, pressure level, real height and density, are written.
    Input: file_name (string)
           observatory (string). Possible values are 'north' or 'south'
           gridstep (float): grid spacing (0.75 degrees for ECMWF and 1.0 degrees for GDAS)
    Output: a txt file with the exact name as the input file name,
            but with .txt as extension in a format that can be read by MARS
    """
    if os.path.exists(os.path.splitext(file_name)[0] + ".txt"):
        logger.critical(
            f"Output file {os.path.splitext(file_name)[0]}.txt already exists. Aborting."
        )
        sys.exit()

    vn, datadict = get_grib_file_data(file_name)

    pressure_level_index = get_plevels(datadict["Temperature"])
    new_pressure_level_index = pressure_level_index[::-1] * int(
        (len(datadict["Temperature"]) / len(pressure_level_index))
    )

    if observatory:
        latitude_obs, longitude_obs = get_observatory_coordinates(observatory)
    else:
        latitude_obs, longitude_obs = lat, lon

    lat_gridpoint, lon_gridpoint = get_closest_gridpoint(latitude_obs, longitude_obs, gridstep)

    # We create the table file and fill it with the information stored in
    # the above variables, plus the height and density computed form them.

    logger.info("creating the txt file containing the selected data...")
    table_file = open(file_name.split(".")[0] + "MAGIC_format.txt", "w")

    for j in np.arange(len(datadict["Temperature"])):

        if (type(datadict["Temperature"][j].values) == float) or (
            len(datadict["Temperature"][j].values) == 1
        ):
            if new_pressure_level_index[j] == 1:
                print(str([0.00] * 34)[1:-1].replace(",", " "), file=table_file)
            if "GeopotentialHeight" in vn:
                height = get_altitude_from_geopotential_height(
                    datadict["GeopotentialHeight"][j].values, latitude_obs
                )
            else:
                height = get_altitude_from_geopotential(
                    datadict["Geopotential"][j].values, latitude_obs
                )

            fields = (
                datadict["Temperature"][j].year - 2000,
                datadict["Temperature"][j].month,
                datadict["Temperature"][j].day,
                datadict["Temperature"][j].hour,
                new_pressure_level_index[j],
                height,
                datadict["Temperature"][j].values,
                datadict["Ucomponentofwind"][j].values,
                datadict["Vcomponentofwind"][j].values,
                RH[j],
            )
            row_str = "{: >6d}{: >6d}{: >6d}{: >6d}{: >6d}{: >10.2f}{: >10.2f}{: >10.2f}{: >10.2f}{: >10.2f}"
            row_str = row_str.format(*fields)
            table_file.write(row_str + "\n")

        else:  # this is just in case the grib file contains more than one grid point
            if new_pressure_level_index[j] == 1:
                print(str([0.00] * 34)[1:-1].replace(",", " "), file=table_file)
            if "GeopotentialHeight" in vn:
                height = np.float(
                    datadict["GeopotentialHeight"][j].values[
                        (datadict["GeopotentialHeight"][j].data()[1] == lat_gridpoint)
                        & (datadict["GeopotentialHeight"][j].data()[2] == lon_gridpoint)
                    ]
                )
            else:
                height = get_altitude_from_geopotential(
                    datadict["Geopotential"][j].values, latitude_obs
                )

            temperature = np.float(
                datadict["Temperature"][j].values[
                    (datadict["Temperature"][j].data()[1] == lat_gridpoint)
                    & (datadict["Temperature"][j].data()[2] == lon_gridpoint)
                ]
            )

            fields = (
                datadict["Temperature"][j].year - 2000,
                datadict["Temperature"][j].month,
                datadict["Temperature"][j].day,
                datadict["Temperature"][j].hour,
                new_pressure_level_index[j],
                height,
                temperature,
                datadict["Ucomponentofwind"][j].values,
                datadict["Vcomponentofwind"][j].values,
                RH[j],
            )
            row_str = "{: >6d}{: >6d}{: >6d}{: >6d}{: >6d}{: >10.2f}{: >10.2f}{: >10.2f}{: >10.2f}{: >10.2f}"
            row_str = row_str.format(*fields)
            table_file.write(row_str + "\n")
    table_file.close()


def read_grib_file_to_magic_from_txt(txt_file):
    """
    :param txt_file:
    :return:
    """
    input_f = open(txt_file, "r")
    output_f = open(txt_file + "MAGIC_format.txt", "w")

    date, year, month, day, hour, mjd, p, T, h, n, U, V, RH = np.loadtxt(
        input_f, usecols=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 12, 13), unpack=True, skiprows=1
    )

    pl = np.unique(p)[::-1]
    pl_index = (np.arange(len(pl)) + 1).tolist()
    new_pl_index = pl_index * int((len(T) / len(pl_index)))
    remaining_index = len(p) - len(new_pl_index)
    new_pl_index = new_pl_index + pl_index[:remaining_index]

    pbar = tqdm(total=len(T))
    for j in np.arange(len(T)):
        pbar.update(1)
        if new_pl_index[j] == 1:
            print(str([0.00] * 34)[1:-1].replace(",", " "), file=output_f)

        fields = (
            int(year[j] - 2000),
            int(month[j]),
            int(day[j]),
            int(hour[j]),
            int(new_pl_index[j]),
            h[j],
            T[j],
            U[j],
            V[j],
            RH[j],
        )
        row_str = (
            "{: >6d}{: >6d}{: >6d}{: >6d}{: >6d}{: >10.2f}{: >10.2f}{: >10.2f}{: >10.2f}{: >10.2f}"
        )
        row_str = row_str.format(*fields)
        output_f.write(row_str + "\n")

    pbar.close()


def run_in_paraller(function_name, file_list, gridstep, observatory=None, lat=None, lon=None):
    """
    Converts in parallel a list of grib files.
    :param function_name: The function to convert the grib files. It can be either read_grib_file_to_magic or
    read_grib_file_to_text
    :type function_name: string
    :param file_list: name of the file containing the list of grib files to process
    :type file_list: string
    :param gridstep: DAS  granularity - distance between the grid points where we have data from DAS
    :type gridstep: float
    :param observatory: The location of the observatory, north or south
    :type observatory: string
    :param lat: The geographical latitude of the observatory
    :type lat: float
    :param lon: The geographical longitude of the observatory
    :type lon: float
    """

    fname = open(file_list)
    line = fname.readline()
    list_of_gribfiles = []
    while line:
        list_of_gribfiles.append(line[:-1])
        line = fname.readline()

    p = multiprocessing.Pool()
    for grib_file in list_of_gribfiles:
        p.map(function_name, [grib_file, gridstep, observatory, lat, lon])

    p.close()
    p.join()


def merge_txt_from_grib(txtfile, output_file="merged_from_grib.txt"):
    """Merges text files created from different grib2 files to one text file"""

    lf = open(txtfile, "r")
    outfile = open(output_file, "w")

    line = lf.readline()
    first = True
    while line:
        datafile = open(line[:-1], "r")
        if first:
            dataline = datafile.readline()
        else:
            datafile.readline()
            dataline = datafile.readline()

        while dataline:
            print(dataline[:-1], file=outfile)
            dataline = datafile.readline()
        first = False
        datafile.close()
        line = lf.readline()
    lf.close()
    outfile.close()


def print_help():
    print("Usage: python grib_utils.py <options>")
    print("Options are:")
    print("        -r         <grib_file_name> <gridstep> <observatory>")
    print("                   note that <gridstep> is 0.75deg for ECMWF data")
    print("                   and 1.0 deg for GDAS data")
    print("        -rmagic    <grib_file_name> <observatory> <gridstep>")
    print("                   note that <gridstep> is 0.75deg for ECMWF data")
    print("                   and 1.0 deg for GDAS data")
    print("        -mjd       <mjd>")
    print("        -date      <yyyy-mm-dd-hh>")
    print("        -merge or -m <list_of_txt_files> <output_name>")
    print(" ")
    print("                   Note: with the -r or -rmagic option, if a txt file")
    print("                   containing a list of grib files is passed instead")
    print("                   of a single grib file, the processing is run in parallel")
    print("                   using a certain number of CPU's")


# if __name__ == "__main__":

parser = argparse.ArgumentParser()
parser.add_argument("-g", "--grib_file", help="the grib file to process")
parser.add_argument(
    "-gridstep",
    help="the gridstep in degrees. If GDAS or GFS data, gridstep=1.0 deg; "
    "If ECMWF data, gridstep=0.75 deg",
    type=float,
)
parser.add_argument(
    "-o",
    "--observatory",
    help="north or south. If no observatory is provided, "
    "then the system asks for the coordinates of interest",
)
parser.add_argument(
    "-c",
    "--coordinates",
    nargs=2,
    help="latitude and longitude of the place " "of interest, in degrees",
    type=float,
)
parser.add_argument(
    "-r",
    action="store_true",
    help="<grib_file_name> <gridstep> <observatory> \n "
    "note that <gridstep> is 0.75deg for ECMWF data \n"
    "and 1.0 deg for GDAS data. If a txt file containing \n "
    "a list of grib files is passed instead of a single \n "
    "grib file, the processing is run in parallel \n"
    "using a certain number of CPUs",
)
parser.add_argument(
    "-rmagic",
    action="store_true",
    help="<grib_file_name> <observatory> <gridstep> \n "
    "note that <gridstep> is 0.75deg for ECMWF data \n "
    "and 1.0 deg for GDAS data.  If a txt file containing \n "
    "a list of grib files is passed instead of a single \n "
    "grib file, the processing is run in parallel \n"
    "using a certain number of CPUs",
)
parser.add_argument(
    "--parallel",
    action="store_true",
    help="if a txt file containing a list of grib files is passed as -g option,"
    "it opens them and extracts information in parallel, one grib file per"
    "CPU",
)
parser.add_argument(
    "-mjd",
    help="if selected, transforms MJD information into date in YYYY MM DD HH format",
    type=float,
)
parser.add_argument(
    "-date", help="if selected, transforms date information into MJD format", type=str
)
parser.add_argument(
    "-m",
    "--merge",
    nargs="+",
    help="followed by a filename containing a list of txt files, \n "
    "it merges them into a single txt file",
)


if __name__ == "__main__":
    args = parser.parse_args()
    print(args)
    if args.grib_file:
        if args.observatory:
            if args.grib_file.lower().endswith((".grib", ".grb", ".txt", ".dat", ".grib2")):
                if args.r and not args.parallel:
                    read_grib_file_to_text(
                        args.grib_file, args.gridstep, observatory=args.observatory
                    )
                elif args.rmagic:
                    read_grib_file_to_magic(
                        args.grib_file, args.gridstep, observatory=args.observatory
                    )
                elif args.r and args.parallel:
                    logger.info("EXECUTING THIS")
                    run_in_paraller(
                        read_grib_file_to_text,
                        args.grib_file,
                        args.gridstep,
                        observatory=args.observatory,
                    )
                elif args.rmagic and args.parallel:
                    run_in_paraller(
                        read_grib_file_to_magic,
                        args.grib_file,
                        args.gridstep,
                        observatory=args.observatory,
                    )
            else:
                logger.critical("file extension not recognized. Exiting")
                sys.exit()
        elif not args.observatory and args.coordinates:
            lat, lon = args.coordinates
            if args.grib_file.lower().endswith((".grib", ".grb", ".txt", ".dat", ".grib2")):
                if args.r:
                    read_grib_file_to_text(args.grib_file, args.gridstep, lat=lat, lon=lon)
                elif args.rmagic:
                    read_grib_file_to_magic(args.grib_file, args.gridstep, lat=lat, lon=lon)
                elif args.r and args.parallel:
                    run_in_paraller(
                        read_grib_file_to_text, args.grib_file, args.gridstep, lat=lat, lon=lon
                    )
                elif args.rmagic and args.parallel:
                    run_in_paraller(
                        read_grib_file_to_magic, args.grib_file, args.gridstep, lat=lat, lon=lon
                    )
            else:
                logger.critical("file extension not recognized. Exiting")
                sys.exit()
        else:
            logger.error(
                "Too many options. Please specify either observatory only, or coordinates only"
            )

    elif args.mjd:
        print(mjd2date(args.mjd))

    elif args.date:
        date = args.date.split("-")
        print(date2mjd(int(date[0]), int(date[1]), int(date[2]), int(date[3])))

    elif args.merge:
        if isinstance(args.merge, list) and len(args.merge) == 2:
            merge_txt_from_grib(args.merge[0], output_file=args.merge[1])
        else:
            merge_txt_from_grib(args.merge[0])



# standard atmosphere thermodynamic values
STD_NUMBER_DENSITY = 2.546899e19  # [cm^-3] molecular number density for standard air conditions
STD_AIR_PRESSURE = 1013.25  # [hPa]   standard air pressure
STD_AIR_TEMPERATURE = 288.15  # [K]     standard air temperature
DENSITY_SCALE_HEIGHT = 9500.0  # [km]    density scale hight for La Palma Winter
STD_RELATIVE_HUMIDITY = 45.9  # [%]     standard air rel. humidity
# atmospheric composition
NITROGEN_RATIO = 0.78084
OXYGEN_RATIO = 0.20946
ARGON_RATIO = 0.00934
GAS_CONSTANT = 8.31451  # gas constant [J/mol/K]
MOLAR_MASS_WATER_VAPOR = 0.018015  # molar mass of water vapor [kg/mol]



from __future__ import division, print_function
from math import pi, cos
import numpy as np
import molecularprofiles.utils.humidity as humidity
from molecularprofiles.utils.meteorological_constants import (
    STD_NUMBER_DENSITY,
    STD_AIR_PRESSURE,
    STD_AIR_TEMPERATURE,
    STD_RELATIVE_HUMIDITY,
    NITROGEN_RATIO,
    OXYGEN_RATIO,
    ARGON_RATIO,
    GAS_CONSTANT,
)


class Rayleigh:
    """
    A small Rayleigh-scattering program based on:
    C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari
    "Improved algorithm for calculations of Rayleigh-scattering optical
    depth in standard atmospheres"
    Applied Optics 44 Nr. 16 (2005) 3320

    The calculation of refractive index is based on:
    P.E. Ciddor, "Refractive index of air: new equations for hte visible and near infrared",
    Applied Optics 35 (1996) 1566

    P.E. Ciddor, "Refractive index of air: 3. The roles of CO2, H20 and refractivity virals",
    Applied Optics 41 (2002) 2292

    The principal King factor formula is based on:
    D.R. Bates, "Rayleigh scattering by air",
    Planet. Space Sci. 32 (1984) 785

    B.A. Bodhaine, N.B. Wood, E.G. Dutton, J.R. Slusser, "On Rayleigh optical depth calculations",
    J. Atmosph. Osceanic Technol. 16 (1999) 1854

    The calculation of the Chandrasekhar phase function is based on:
    S. Chandrasekhar, Radiative Transfer, Dover Publications, 1960.

    E.J. McCartney, "Optics of the Atmosphere. Scattering by Molecules and Particles"
    Wiley & Sons, New York, 1977.

    Adapted from MRayleigh, written by Markus Gaug <markus.gaug@uab.cat>, 04/2013
    .. moduleauthor:: Scott Griffiths <sgriffiths@ifae.es>
    """

    def __init__(
        self,
        wavelength,
        co2_bkg,
        pressure=STD_AIR_PRESSURE,
        temperature=STD_AIR_TEMPERATURE,
        relative_humidity=STD_RELATIVE_HUMIDITY,
    ):
        """
        Constructor for Rayleigh-scattering class.

        Args:
            wavelength : wavelength of light [nm]
            pressure  : pressure [hPa]
            temperature  : temperature [K]
            relative_humidity : relative humidity [%]
            co2_bkg  : CO2 concentration [ppmv]
        """
        # check inputs for bad values
        if wavelength < 200 or wavelength > 4000:
            raise ValueError("Wavelength range only from 200 nm - 4 micrometer allowed.")
        if pressure < 0 or pressure > 1400:
            raise ValueError("Pressure only in range 0 - 1400 hPa allowed.")
        if temperature < 180 or temperature > 373.15:
            raise ValueError("Temperatures only in range 200 - 373 K allowed.")
        if relative_humidity < 0 or relative_humidity > 100:
            raise ValueError("Relative humity must lie between 0 - 100.")
        if co2_bkg < 200 or co2_bkg > 1000:
            raise ValueError("CO2 concentrations only in range 200 - 1000 ppmv allowed.")

        self.wavelength = wavelength  # [nm]    wavelenght of light
        self.pressure = pressure  # [hPa]   air pressure
        self.temperature = temperature  # [K]     air temperature
        self.relative_humidity = relative_humidity  # [%]     relative humidity of air
        self.co2_bkg = co2_bkg  # [ppmv]  CO2 concentration of air

        self.calculate_sigma()  # [cm^-2] single molecule scattering coefficient
        self.calculate_beta()  # [km^-1] volume scattering coefficient

    def calculate_sigma(self):
        """
        Calculate the total Rayleigh scattering cross section per moleculem, sigma,
        in units of cm^-2.
        """
        wavelength_cm = self.wavelength * 1e-7
        wavelength_4 = pow(wavelength_cm, 4)

        # calculate the refractive index of air
        refractive_index = self.calculate_n()
        refractive_index_sqrd = refractive_index * refractive_index

        # calculate molecular number density in units of cm^-3 (Tomasi eq. 3)
        number_density = (
            STD_NUMBER_DENSITY
            * self.pressure
            / STD_AIR_PRESSURE
            * STD_AIR_TEMPERATURE
            / self.temperature
        )
        number_density_sqrd = number_density * number_density

        pi3 = pow(pi, 3)
        king_factor = self.king()
        self.sigma = (
            24
            * pi3
            * (refractive_index_sqrd - 1) ** 2
            / (wavelength_4 * number_density_sqrd * (refractive_index_sqrd + 2) ** 2)
            * king_factor
        )  # units of cm^-2
        return self.sigma

    def calculate_beta(self):
        """
        Calculates the monochromatic volume coefficient for the total molecular
        scattering in cloudless air, beta, in units of km^-1
        """
        # calculate molecular number density in units of cm^-3 (Tomasi eq. 3)
        number_density = (
            STD_NUMBER_DENSITY
            * self.pressure
            / STD_AIR_PRESSURE
            * STD_AIR_TEMPERATURE
            / self.temperature
        )

        try:  # check that sigma has been calculated
            self.sigma
        except NameError:
            self.calculate_sigma()

        # calculate beta in units of km^-1 (Tomasi eq. 2)
        self.beta = 1e5 * number_density * self.sigma
        return self.beta

    def calculate_n(self):
        """
        Ciddor formula for calculation of refractive index in moist air.
        The obtained refractive index is precise to 1e-7.

        Cross-checked with:
        http://emtoolbox.nist.gov/Wavelength/Documentation.asp#IndexofRefractionofAir

        Returns:
            refractive_index : index of refraction of moist air
        """
        wavelength_mum = pow(self.wavelength / 1000, -2)  # convert wavelength to micrometers

        # refractive index of standard dry air (e = 0) according to Ciddor, with 450 ppmv CO2
        refractive_index_dry = (
            1e-8 * (5792105 / (238.0185 - wavelength_mum) + 167917 / (57.362 - wavelength_mum)) + 1
        )  # Tomasi eq. 17

        # refractive index of dry air at standard p and T, for given C (e = 0)
        refractive_index_dry_std_air = (1 + 0.534e-6 * (self.co2_bkg - 450)) * (
            refractive_index_dry - 1
        ) + 1  # Tomasi eq. 18

        # refractive index of pure water vapor at standard T and e
        # (T* = 293.15 K = 20 C, and e* = 1333 Pa)
        refractive_index_water_vapour = (
            1.022e-8
            * (
                295.235
                + 2.6422 * wavelength_mum
                - 0.032380 * wavelength_mum * wavelength_mum
                + 0.004028 * wavelength_mum * wavelength_mum * wavelength_mum
            )
            + 1
        )  # Tomasi eq. 19

        # calculate the respective densities (see Tomasi et al., pp. 3325 ff)
        molar_mass_dry_air = 1e-3 * (
            28.9635 + 12.011e-6 * (self.co2_bkg - 400)
        )  # molar mass of dry air [kg/mol]
        molar_mass_water_vapour = 0.018015  # molar mass of water vapor [kg/mol]
        molar_fraction_water_vapour = humidity.molar_fraction_water_vapor(
            self.pressure, self.temperature, self.relative_humidity
        )  # molar fraction of water vapor in moist air
        compressibility_dry_air = humidity.compressibility(
            STD_AIR_PRESSURE, STD_AIR_TEMPERATURE, 0
        )  # compressibility of dry air
        compressibility_water_vapour = humidity.compressibility(
            13.33, 293.15, 1
        )  # compressibility of pure water vapor
        compressibility_moist_air = humidity.compressibility(
            self.pressure, self.temperature, molar_fraction_water_vapour
        )  # compressibility of moist air

        # density of dry air at standard p and T
        dry_air_density_stdpt = humidity.density_moist_air(
            STD_AIR_PRESSURE, STD_AIR_TEMPERATURE, compressibility_dry_air, 0, self.co2_bkg
        )

        # density of pure water vapor at at standard T and e (T* = 293.15 K = 20 C, and e* = 1333 Pa)
        water_vapour_density_stdpt = humidity.density_moist_air(
            13.33, 293.15, compressibility_water_vapour, 1, self.co2_bkg
        )

        # density of the dry component of moist air
        density_dry_comp_moist_air = (
            (100 * self.pressure)
            * molar_mass_dry_air
            * (1 - molar_fraction_water_vapour)
            / (compressibility_moist_air * GAS_CONSTANT * self.temperature)
        )

        # density of the water vapor component of moist air
        density_water_vapour_moist_air = (
            (100 * self.pressure)
            * molar_mass_water_vapour
            * molar_fraction_water_vapour
            / (compressibility_moist_air * GAS_CONSTANT * self.temperature)
        )

        refractive_index = (
            1
            + (density_dry_comp_moist_air / dry_air_density_stdpt)
            * (refractive_index_dry_std_air - 1)
            + (density_water_vapour_moist_air / water_vapour_density_stdpt)
            * (refractive_index_water_vapour - 1)
        )  # Ciddor eq. 5, Tomasi eq. 11
        return refractive_index

    def king(self):
        """
        Calculates the current best estimate of the King factor of moist air.

        The King factor is used to take into account effects due to the anisotropic
        properties of air molecules since anisotropic molecules scatter more radiation
        at 90 degrees scattering angles than isotropic molecules with the same index
        of refraction.

        Precision not stated in Tomasi et al., but probably better than 1e-4.
        Effects of relative_humidity are of the order of several times 1e-4.

        Returns:
            king_factor : King factor [dimensionless]
        """
        wavelength_mum = pow(self.wavelength / 1000, -2)  # convert to micrometers
        water_vapour_partial_press = humidity.partial_pressure_water_vapor(
            self.temperature, self.relative_humidity
        )  # water vapor partial pressure [hPa]

        king_factor_n2 = 1.034 + 3.17e-4 * wavelength_mum  # partial King factor for N2 molecules
        king_factor_o2 = (
            1.096 + 1.385e-3 * wavelength_mum + 1.448e-4 * wavelength_mum * wavelength_mum
        )  # partial King factor for O2 molecules
        king_factor_ar = 1.00  # partial King factor for Ar molecules
        king_factor_co2 = 1.15  # partial King factor for CO2 molecules
        king_factor_wv = 1.001  # partial King factor for water vapor

        co2_ratio = 1e-6 * self.co2_bkg  # CO2
        water_vapour_ratio = water_vapour_partial_press / self.pressure  # water vapor mixing ratio

        king_factor = (
            NITROGEN_RATIO * king_factor_n2
            + OXYGEN_RATIO * king_factor_o2
            + ARGON_RATIO * king_factor_ar
            + co2_ratio * king_factor_co2
            + water_vapour_ratio * king_factor_wv
        ) / (
            NITROGEN_RATIO + OXYGEN_RATIO + ARGON_RATIO + co2_ratio + water_vapour_ratio
        )  # Tomasi eq. 22
        return king_factor

    def phase_function(self, angle):
        """
        Calculates the Chandrasekhar phase function.

        Args:
            angle : scattering angle in radians

        Returns:
            Chandrasekhar phase function for scattering of natural light
        """
        rho = self.depolarization()

        # need to solve Chandrasekhar eq. 254 for gamma as a function of rho
        f_1 = (2 + 2 * rho) / (2 + rho)
        f_2 = (1 - rho) / (1 + rho)
        return 0.75 * f_1 * (1 + f_2 * pow(cos(angle), 2))  # Chandrasekhar eq. 255

    def depolarization(self):
        """
        Current best estimate of the depolarization factor of moist air.

        Precision not stated in Tomasi et al., but probably better than 1e-4.
        Effects of relative_humidity are of the order of several times 1e-4.
        """
        king_factor = self.king()
        return 6 * (king_factor - 1) / (3 + 7 * king_factor)  # Tomasi eq. 5, solved for rho

    # TODO: where does this come from? Needs unit test
    def dbeta_domega(self, angle):
        """
        Get the back-scattering coefficient for a given scattering angle.

        Args:
            angle : scattering angle in radians

        Returns:
            back-scattering coefficient in units of km^-1
        """
        try:  # check that beta has been calculated
            self.beta
        except NameError:
            self.calculate_beta()

        phase_func = self.phase_function(angle)
        return phase_func * self.beta / (4 * pi)

    def print_params(self):
        """Prints Rayleigh scattering parameters."""
        print(f"Wavelength:              {self.wavelength} nm")
        print(f"Air Pressure:            {self.pressure} hPa")
        print(f"Air Temperature:         {self.temperature} K")
        print(f"Rel. Humidity:           {self.relative_humidity} %")
        print(f"CO2 concentration:       {self.co2_bkg} ppmv")
        print(f"Refractive Index:        {self.calculate_n()}")
        print(f"King Factor:             {self.king()}")
        print(f"Depolarization:          {self.depolarization()}")
        print(f"Mol. cross section:      {self.sigma} cm^-2")
        print(f"Volume scattering coeff: {self.beta} km^-1")



"""
Calculates the properties of moist air, including:
* density
* compressibility
* enhancement factor
* saturation vapor pressure
* molar fraction from relative humidity
* partial pressure from relative humidity

Adapted from MHumidity, written by Markus Gaug <markus.gaug@uab.cat>, 04/2013
"""

from math import exp, sqrt, log10
import numpy as np
from molecularprofiles.utils.meteorological_constants import GAS_CONSTANT, MOLAR_MASS_WATER_VAPOR


def compressibility(pressure, temp, x_w):
    """
    Calculates the compressibility of moist air, according to:
    R.S. Davis, "Equation for the determination of the density of moist air"
    Metrologia, 29 (1992) 67-70

    See also Eq. 16 in:
    C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari
    "Improved algorithm for calculations of Rayleigh-scattering optical depth
    in standard atmospheres", Applied Optics 44 Nr. 16 (2005) 3320

    Args:
        pressure  : pressure in hPa
        temp  : temperature in K
        x_w : molar fraction of water vapor
    Returns:
        compressibility of moist air (dimensionless constant, 0 < Z < 1)
    """
    temp_celsius = temp - 273.15  # temperature in deg. C
    pressure_temp_ratio = (
        100 * pressure / temp
    )  # ratio of pressure to temperature, with pressure in Pascals
    a_0 = 1.58123e-6  # K Pa^-1
    a_1 = -2.9331e-8  # Pa^-1
    a_2 = 1.1043e-10  # K^-1 Pa^-1
    b_0 = 5.707e-6  # K Pa^-1
    b_1 = -2.051e-8  # Pa^-1
    c_0 = 1.9898e-4  # K Pa^-1
    c_1 = -2.376e-6  # Pa^-1
    d_0 = 1.83e-11  # K^2 Pa^-2
    d_1 = -7.65e-9  # K^2 Pa^-2
    moist_air_compressibility = (
        1
        - pressure_temp_ratio
        * (
            a_0
            + a_1 * temp_celsius
            + a_2 * temp_celsius * temp_celsius
            + b_0 * x_w
            + b_1 * x_w * temp_celsius
            + c_0 * x_w * x_w
            + c_1 * x_w * x_w * temp_celsius
        )
        + pressure_temp_ratio * pressure_temp_ratio * (d_0 + d_1 * x_w * x_w)
    )
    return moist_air_compressibility


def enhancement_factor(pressure, temp):
    """
    Calculates the enhancement factor of water vapor in air.

    Calculated according to Eq. 14 of:
    C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari
    "Improved algorithm for calculations of Rayleigh-scattering optical depth
    in standard atmospheres", Applied Optics 44 Nr. 16 (2005) 3320

    Args:
        p : pressure in Pa
        T : temperature in K
    Returns:
        enhancement factor (dimensionless constant)
    """
    temp_celsius = temp - 273.15  # temparture in deg. C
    factor = 1.00062 + 3.14e-8 * pressure + 5.6e-7 * pow(temp_celsius, 2)
    return factor


# TODO: This function needs a simple unit test to check which function is called
def saturation_vapor_pressure(temp):
    """Define which function will be called depending on temperature"""
    if temp > 273.15:
        return saturation_vapor_pressure_davis(temp)
    else:
        return saturation_vapor_pressure_goff_gratch(temp)


def saturation_vapor_pressure_davis(temp):
    """
    Calculates the vapor pressure at saturation, according to:
    R.S. Davis, "Equation for the determination of the density of moist air"
    Metrologia, 29 (1992) 67-70

    See also Eq. 15 in:
    C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari
    "Improved algorithm for calculations of Rayleigh-scattering optical depth
    in standard atmospheres", Applied Optics 44 Nr. 16 (2005) 3320

    Args:
        temp : temperature in K
    Returns:
        saturation vapor pressure in hPa
    """
    psv_davis = np.exp(
        1.2378847e-5 * temp * temp - 1.9121316e-2 * temp + 33.93711047 - 6343.1645 / temp
    )
    return psv_davis / 100  # divide by 100 for hPa


# TODO: Goff-Gratch doesn't seem to be as good as Buck, or maybe Wexler
def saturation_vapor_pressure_goff_gratch(temp):
    """
    Calculates the vapor pressure at saturation, according to:
    Smithsonian Tables (1984); after Goff and Gratch (1946).
    See here: http://cires.colorado.edu/~voemel/vp.html

    This equation is recommended for temperatures below 0 deg. C.

    Args:
        temp : temperature in K
    Returns:
        saturation vapor pressure in hPa
    """
    theta = 373.16 / temp  # ratio of steam point (100 deg C) to temperature
    c = [
        -7.90298 * (theta - 1),
        5.02808 * log10(theta),
        -1.3816e-7 * (pow(10, 11.344 * (1 - 1 / theta)) - 1),
        8.1328e-3 * (pow(10, -3.49149 * (theta - 1)) - 1),
        log10(1013.246),
    ]
    log10_ew = np.sum(c)
    psv_goff_gratch = pow(10, log10_ew)
    return psv_goff_gratch


# TODO: Check this function and write unit test
def saturation_vapor_pressure_over_water(temp):
    """
    Calculates the vapor pressure at saturation over water, according to IAPWS:
    International Association for the Properties of Water and Steam,
    Peter H. Huang, "New equations for water vapor pressure in the temperature
    range -100 deg. C to 100 deg. C for use with the 1997 NIST/ASME steam tables"
    Papers and abstracts from the third international symposium on humidity and
    moisture, Vol. 1, p. 69-76, National Physical Laboratory, Teddington,
    Middlesex, UK, April 1998.
    See also: http://cires.colorado.edu/~voemel/vp.html

    Args:
        temp : temperature in K
    Returns:
        saturation vapor pressure in hPa
    """
    omega = temp - 2.38555575678e-01 / (temp - 6.50175348448e02)
    omega2 = omega * omega
    A = omega2 + 1.16705214528e03 * omega - 7.24213167032e05
    B = -1.70738469401e01 * omega2 + 1.20208247025e04 * omega - 3.23255503223e06
    C = 1.49151086135e01 * omega2 - 4.82326573616e03 * omega + 4.05113405421e05
    water_psv = -B + sqrt(B * B - 4 * A * C)
    return 1e4 * pow(2 * C / water_psv, 4)


# TODO: Check this function and write unit test
def saturation_vapor_pressure_over_ice(temp):
    """
    Calculates the vapor pressure at saturation over ice, according to IAPWS:
    International Association for the Properties of Water and Steam,
    Peter H. Huang, "New equations for water vapor pressure in the temperature
    range -100 deg. C to 100 deg. C for use with the 1997 NIST/ASME steam tables"
    Papers and abstracts from the third international symposium on humidity and
    moisture, Vol. 1, p. 69-76, National Physical Laboratory, Teddington,
    Middlesex, UK, April 1998.
    See also: http://cires.colorado.edu/~voemel/vp.html

    Args:
        temp : temperature in K
    Returns:
        saturation vapor pressure in hPa
    """
    theta = temp / 273.16
    ice_psv = -13.928169 * (1 - pow(theta, -1.5)) + 34.7078238 * (1 - pow(theta, -1.25))
    return 6.11657 * exp(ice_psv)


def molar_fraction_water_vapor(pressure, temp, relative_humidity):
    """
    Calculates the molar fraction of water vapor in moist air.

    See the text above Eq. 14 of:
    C. Tomasi, V. Vitale, B. Petkov, A. Lupi, A. Cacciari
    "Improved algorithm for calculations of Rayleigh-scattering optical depth
    in standard atmospheres", Applied Optics 44 Nr. 16 (2005) 3320

    Args:
        p  : pressure in hPa
        temp  : temperature in K
        relative_humidity : relative humidity in percent
    Returns:
        x_w : molar fraction of water vapor in moist air (dimensionless)
    """
    factor = enhancement_factor(100 * pressure, temp)
    psv = saturation_vapor_pressure(temp)
    x_w = factor * relative_humidity / 100 * psv / pressure  # x_w = f * h * E(temp)/pressure
    return x_w


def density_moist_air(pressure, temp, moist_air_compressibility, x_w, co2_bkg):
    """
    Density equation of moist air, according to:
    R.S. Davis, "Equation for the determination of the density of moist air"
    Metrologia, 29 (1992) 67-70

    Args:
        pressure  : pressure in hPa (beware, different unit than in Davis!)
        temp  : temperature in Kelvin
        moist_air_compressibility  : compressibility (see compressibility() in this module)
        x_w : molar fraction of water vapor
        co2_bkg  : CO2 volume concentration in ppmv (different unit than in Davis!)
    Returns:
        density of moist air (kg m^-3)
    """
    pressure *= 100  # convert pressure to Pa
    R = GAS_CONSTANT
    m_w = MOLAR_MASS_WATER_VAPOR
    m_a = 1e-3 * (28.9635 + 12.011e-6 * (co2_bkg - 400))  # molar mass of dry air [kg/mol]
    rho = (
        pressure * m_a / (moist_air_compressibility * R * temp) * (1 - x_w * (1 - m_w / m_a))
    )  # Tomasi eq. 12
    return rho


# TODO: This function has no unit test
def partial_pressure_water_vapor(temp, relative_humidity):
    """
    Calculates the partial pressure of water vapor in the air.

    Args:
        temp  : temperature in K
        relative_humidity : relative humidity in percent
    Returns:
        ew : water vapor partial pressure in hPa
    """
    # water vapor pressure: e = h * E(temp)
    e_w = (relative_humidity / 100) * saturation_vapor_pressure_davis(temp)
    return e_w



"""
Module containing functionality allows to get observatory's coordinates
and select meteorological data depending on the season for the South and North hemisphere

"""

import numpy as np
import os
import logging
import logging.config

ROOTDIR = os.path.dirname(os.path.abspath(__file__))
log_config_file = f"{ROOTDIR}/mdps_log.conf"
logging.config.fileConfig(fname=log_config_file, disable_existing_loggers=False)
logger = logging.getLogger(__name__)


def ddmmss2deg(degrees, minutes, seconds):
    """
    This function computes the value, in degrees, of an angle expressed in degrees, minutes, seconds
    Input: (integers) degrees, minutes, seconds
    Output: (float) angle in degrees
    """

    if degrees > 0.0:
        angle = degrees + (minutes / 60.0) + (seconds / 3600.0)
    else:
        angle = degrees - (minutes / 60.0) - (seconds / 3600.0)
    return angle


def get_observatory_coordinates(observatory):
    """Give latitude, longitude of the observatory"""
    if observatory == "north":
        latitude = ddmmss2deg(28, 45, 42.462)
        longitude = ddmmss2deg(-17, 53, 26.525)
        return latitude, longitude
    elif observatory == "south":
        latitude = ddmmss2deg(-24, 40, 24.8448)
        longitude = ddmmss2deg(-70, 18, 58.4712)
        return latitude, longitude
    else:
        logger.error("Unknown observatory!")
        latitude = np.float(input("Observatory latitude (in degrees):"))
        longitude = np.float(input("Observatory longitude (in degrees):"))
        return latitude, longitude


def get_winter_months():
    """Return winter months for the North hemisphere"""
    return [1, 2, 3, 4, 11, 12]


def get_summer_months():
    """Return summer months for the North hemisphere"""
    return [6, 7, 8, 9, 10]


def get_all_months():
    """Return all months for any hemisphere"""
    return [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]


def get_intermediate_months():
    """Return spring/autumn months for the North hemisphere"""
    return [5, 6, 10, 11]


def get_epoch(epoch):
    """Return the season for the North hemisphere"""
    valid_epochs = {
        "winter": get_winter_months(),
        "summer": get_summer_months(),
        "intermediate": get_intermediate_months(),
        "all": get_all_months(),
    }
    return valid_epochs[epoch]


def get_south_winter_months():
    """Return winter months for the South hemisphere"""
    return [5, 6, 7, 8, 9, 10]


def get_south_summer_months():
    """Return summer months for the South hemisphere"""
    return [1, 2, 3, 4, 5, 10, 11, 12]


def get_south_epoch(epoch):
    """Return the season for the South hemisphere"""
    valid_epochs = {
        "winter": get_south_winter_months(),
        "summer": get_south_summer_months(),
        "all": get_all_months(),
    }
    return valid_epochs[epoch]


def select_new_epochs_dataframe_north(dataframe_atmo_profile, epoch_text):
    """
    Select part of dataframe rows for a given season (North)
    The selection is date-based
    """

    epoch = get_epoch(epoch_text)

    if epoch_text == "winter":
        condition = (
            (dataframe_atmo_profile.month == epoch[0])
            | (dataframe_atmo_profile.month == epoch[1])
            | (dataframe_atmo_profile.month == epoch[2])
            | (dataframe_atmo_profile.month == epoch[3])
            | ((dataframe_atmo_profile.month == epoch[4]) & (dataframe_atmo_profile.day > 15))
            | (dataframe_atmo_profile.month == epoch[5])
        )
        new_dataframe_atmo_profile = dataframe_atmo_profile[condition]

    elif epoch_text == "summer":
        condition = (
            ((dataframe_atmo_profile.month == epoch[0]) & (dataframe_atmo_profile.day > 20))
            | (dataframe_atmo_profile.month == epoch[1])
            | (dataframe_atmo_profile.month == epoch[2])
            | (dataframe_atmo_profile.month == epoch[3])
        )
        new_dataframe_atmo_profile = dataframe_atmo_profile[condition]

    elif epoch_text == "intermediate":
        condition = (
            (dataframe_atmo_profile.month == epoch[0])
            | ((dataframe_atmo_profile.month == epoch[1]) & (dataframe_atmo_profile.day <= 20))
            | ((dataframe_atmo_profile.month == epoch[2]) & (dataframe_atmo_profile.day > 5))
            | ((dataframe_atmo_profile.month == epoch[3]) & (dataframe_atmo_profile.day <= 15))
        )
        new_dataframe_atmo_profile = dataframe_atmo_profile[condition]

    return new_dataframe_atmo_profile


def select_new_epochs_dataframe_south(dataframe_atmo_profile, epoch_text):
    """
    Select part of dataframe rows for a given season (South)
    The selection is date-based
    """

    epoch = get_south_epoch(epoch_text)

    if epoch_text == "summer":
        condition = (
            (dataframe_atmo_profile.month == epoch[0])
            | (dataframe_atmo_profile.month == epoch[1])
            | (dataframe_atmo_profile.month == epoch[2])
            | (dataframe_atmo_profile.month == epoch[3])
            | ((dataframe_atmo_profile.month == epoch[4]) & (dataframe_atmo_profile.day < 15))
            | ((dataframe_atmo_profile.month == epoch[5]) & (dataframe_atmo_profile.day > 15))
            | (dataframe_atmo_profile.month == epoch[6])
            | (dataframe_atmo_profile.month == epoch[7])
        )
    elif epoch_text == "winter":
        condition = (
            ((dataframe_atmo_profile.month == epoch[0]) & (dataframe_atmo_profile.day > 15))
            | (dataframe_atmo_profile.month == epoch[1])
            | (dataframe_atmo_profile.month == epoch[2])
            | (dataframe_atmo_profile.month == epoch[3])
            | (dataframe_atmo_profile.month == epoch[4])
            | ((dataframe_atmo_profile.month == epoch[5]) & (dataframe_atmo_profile.day < 15))
        )
    new_dataframe_atmo_profile = dataframe_atmo_profile[condition]
    return new_dataframe_atmo_profile


def select_new_epochs_dataframe_density_north(dataframe_atmo_profile, epoch_text):
    """
    Select part of dataframe rows for a given season (North)
    The selection is based on the value of the atmospheric molecular density
    """

    if epoch_text == "summer":
        condition = dataframe_atmo_profile[
            (dataframe_atmo_profile.n_exp > 0.88) & (dataframe_atmo_profile.P == 125)
        ]
    elif epoch_text == "winter":
        condition = dataframe_atmo_profile[
            (dataframe_atmo_profile.n_exp < 0.8375) & (dataframe_atmo_profile.P == 125)
        ]
    elif epoch_text == "intermediate":
        condition = dataframe_atmo_profile[
            (dataframe_atmo_profile.n_exp < 0.88)
            & (dataframe_atmo_profile.n_exp > 0.8375)
            & (dataframe_atmo_profile.P == 125)
        ]
    new_dataframe_atmo_profile = dataframe_atmo_profile[
        dataframe_atmo_profile.MJD.isin(condition.MJD)
    ]
    return new_dataframe_atmo_profile


def select_new_epochs_dataframe_density_south(dataframe_atmo_profile, epoch_text):
    """
    Select part of dataframe rows for a given season (South)
    The selection is based on the value of the atmospheric molecular density
    """

    if epoch_text == "summer":
        condition = dataframe_atmo_profile[
            (dataframe_atmo_profile.n_exp > 0.88) & (dataframe_atmo_profile.P == 125)
        ]
    elif epoch_text == "winter":
        condition = dataframe_atmo_profile[
            (dataframe_atmo_profile.n_exp < 0.88) & (dataframe_atmo_profile.P == 125)
        ]
    new_dataframe_atmo_profile = dataframe_atmo_profile[
        dataframe_atmo_profile.MJD.isin(condition.MJD)
    ]
    return new_dataframe_atmo_profile


def select_epoch(file, epoch_text):
    """
    Select part of the txt converted grib2 data for a given season.
    The selection is based on the value of the atmospheric molecular density
    """

    global months, month
    logger.info("loading and selecting data")
    file = open(file)
    (
        date,
        year,
        month,
        day,
        hour,
        mjd,
        pressure,
        temperature,
        height,
        n_density,
        u_component,
        v_component,
        relative_humidity,
    ) = np.loadtxt(
        file, usecols=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 12, 13), unpack=True, skiprows=1
    )

    epoch = get_epoch(epoch_text)

    if epoch_text == "all":
        condition = month < 13
    elif epoch_text == "winter":
        condition = (
            (month == epoch[0]) | (month == epoch[1]) | (month == epoch[2]) | (month == epoch[3])
        )
    elif epoch_text == "summer":
        condition = (month == epoch[0]) | (month == epoch[1]) | (month == epoch[2])
    elif epoch_text == "intermediate":
        condition = (
            (month == epoch[0])
            | (month == epoch[1])
            | (month == epoch[2])
            | (month == epoch[3])
            | (month == epoch[4])
        )
    date = date[condition]
    modified_julian_date = mjd[condition]
    height = height[condition]
    number_density = n_density[condition]
    pressure = pressure[condition]
    temperature = temperature[condition]
    u_wind_component = u_component[condition]
    v_wind_component = v_component[condition]
    relative_humidity = relative_humidity[condition]

    # print(date, year, month, day, hour, modified_julian_date, pressure, height, number_density, temperature, u_wind_component, v_wind_component, relative_humidity)
    return (
        date,
        year,
        month,
        day,
        hour,
        modified_julian_date,
        pressure,
        height,
        number_density,
        temperature,
        u_wind_component,
        v_wind_component,
        relative_humidity,
    )


def get_closest_gridpoint(lat, lon, gridstep):
    """
    :param lat: float
    :param lon: float
    :param gridstep: float. Step in which the grid is divided (0.75 degrees for ECMWF data
    and 1.0 degrees for GDAS data)
    :return: nearest grid point
    """
    step = gridstep  # grid step in degrees
    lons_grid = np.zeros(int(360 / step) + 1)
    lats_grid = np.zeros(int(180 / step) + 1)

    # getting the grid points:
    for i in range(len(lons_grid)):
        # lons_grid[i] = step * i
        lons_grid[i] = -180 + step * i
    for i in range(len(lats_grid)):
        lats_grid[i] = -90 + step * i

    logger.info(f"Latidude of interest is {lat:5.2f}")
    # logger.info(f"Longitude of interest is {(lon % 360):5.2f}")
    logger.info(f"Longitude of interest is {lon:5.2f}")
    nearest_lat = find_nearest(lats_grid, lat)
    logger.info(f"nearest latitude is {nearest_lat:4.1f}")
    # nearest_lon = find_nearest(lons_grid, lon % 360)
    nearest_lon = find_nearest(lons_grid, lon)
    logger.info(f"nearest longitude is {nearest_lon:5.1f}")
    return nearest_lat, nearest_lon


def find_nearest(grid_position, value):
    """Function to find the nearest grid position to a given latitude or longitude"""
    return grid_position[np.abs(grid_position - value).argmin()]
